# Python code for performing a parameter sweep
# of conductance shift as a function of B-field

# Toby Jacobson, 2018-05-16
# Sandia National Laboratories

# Import modules:
from numpy import *
import subprocess
from time import time
from scipy.interpolate import interp1d
from scipy.optimize import minimize
import functools
import pickle
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
numCores=comm.Get_size()
rankList = range(numCores)

printCheck = (rank==0)
if printCheck:
    print 'Number of cores requested: '+str(numCores)
   
##### Run parameters: ################################
nHoleDensities = 15 # Number of hole densities to consider (coincides with number of experimental measurements)
nTauPhi = 64 # Number of samples to take for the parameter tauPhi
nTauSO = 64 # Number of samples to take for the parameter tauSO
# nTauPhi = 4 # Number of samples to take for the parameter tauPhi
# nTauSO = 4 # Number of samples to take for the parameter tauSO
raTauPhiRange = 1e-12*logspace(0.5,3,nTauPhi) # Range of tauPhi values to compute [s]
raTauSORange = 1e-12*logspace(-1,2,nTauSO) # Range of tauSO values to compute [s]
rLowerBLimit = 0.1 # Lower B-field limit [mT]
rUpperBLimit = 10 # Upper B-field limit [mT]
nBValues = 51 # Number of B-field values to evaluate for the range [rLowerBLimit,rUpperBLimit]
raB = linspace(rLowerBLimit*1e-3,rUpperBLimit*1e-3,nBValues) # Magnetic fields to consider [T]
nNCutoff = 100 # N index at which to revert to asymptotic form
nQuadPoints = 10000 # Number of integration quadrature points
verbosity = False
fOutputPrefix = 'Out_WAL_calc_'
fOutputSuffix = '.pkl'
 
##### Identify the parameter range for this core to compute: #####
iaHoleDensities = range(nHoleDensities) # Each core will compute conductance shifts for all 15 measured hole densities
iaTauPhiTauSO = []
for iTauPhi in xrange(nTauPhi):
    for iTauSO in xrange(nTauSO):
        iaTauPhiTauSO += [[iTauPhi,iTauSO]]
iaTauPhiTauSO = array(iaTauPhiTauSO)
assert len(iaTauPhiTauSO)%numCores == 0, 'Incommensurate number of assigned cores and data points to compute'
nTermsPerCore = len(iaTauPhiTauSO)/numCores
iaTauPhiTauSOToCompute = iaTauPhiTauSO[rank*nTermsPerCore:(rank+1)*nTermsPerCore] # [iTauPhi, iTauSO] for this core to compute
##################################################################
def sigmaTerms(nNCutoff,rl,rlB,rlTilde,rOmegaTau,nQuadPoints,verbose=False):
    '''
    Python wrapper for calling C code that evaluates the conductance corrections
    '''
    sFilename = './conductanceShift'
    args = [sFilename,str(rlB/rl),str(rl/rlTilde),str(rOmegaTau),str(nNCutoff),str(nQuadPoints)]
    if verbose:
        print('(lB/lr, l/lTilde, OmegaTau, NCutoff, nQuadPoints) = '+str(rlB/rl)+', '+str(rl/rlTilde)+', '+str(rOmegaTau)+', '+str(nNCutoff)+', '+str(nQuadPoints))
    proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    soutput = str.split(stdout)
    routput = [float(term) for term in soutput]
    return routput[1]

def evalModelPrediction(raB,rBtr,rTauTauPhi,rOmegaTau,nNCutoff,nQuadPoints,nNumSamples=51,verbose=False):
    '''
    Given the fit parameters,
    raB = Array of B-fields considered [T]
    rBtr = transport field [T]
    rTauTauPhi = tau/tau_{\phi} [unitless]
    rOmegaTau = SO strength [unitless]
    evaluate the predicted shift in conductance in units of [e^2/h]
    over a range spanning B in [-30,30] mT
    '''
    ralB = sqrt(rhbar/(reConst*raB)) # Magnetic lengths [m]
    rl = sqrt(rhbar/(2*reConst*rBtr))
    rlTilde = rl/(1+rTauTauPhi)
    rasig = [] # Conductance shifts [e^2/h]
    time0 = time()
    rasig += [(2*pi*pi)*sigmaTerms(nNCutoff,rl,rlB,rlTilde,rOmegaTau,nQuadPoints, verbose=verbose) for rlB in ralB] # Convert to units of [e^2/(\pi h)]
    raBReflected = concatenate((-array(raB[::-1]),raB)) # Reflect symmetrically across B=0
    rasigReflected = concatenate((array(rasig[::-1]),rasig))
    time1 = time()
    if verbose:
        print('Elapsed time: '+str(time1-time0)+' s.')
    return raBReflected, rasigReflected
##################################################################
sMobilityVsDensity = 'BS460muvsn.txt' # Filename of data in two-column format: [density] [mobility], for a total of 15 densities
sWALData = 'BS460WAL.txt' # Filename of WAL data in 30-column format for all 15 densities:
                          # [B-field] [delta sigma (in units of [e^2/(pi*h)])]
# Hole mobility, in units of [1]/[cm^2] on the x-axis and [cm^2]/([V][s]) on the y-axis:
aMobilityVsDensity = loadtxt(sMobilityVsDensity)
# Conductance shift data, in units of [T] on the x-axis and [e^2/(pi h)] on the y-axis: 
aWALData = loadtxt(sWALData)

# Relevant physical constants:
reConst = 1.60217662e-19 # Electron charge [C]
rm0Const = 9.10938356e-31 # Electron bare mass [kg]
rhbar = 1.0545718e-34 # hbar [J][s]

# Evaluate derived quantities:
raTauTR = (0.08*rm0Const/reConst)*1e-4*aMobilityVsDensity[:,1] # Mean free time [s]
raDiffusionD = (pi*rhbar**2/(reConst*0.08*rm0Const))*aMobilityVsDensity[:,1]*aMobilityVsDensity[:,0] # Diffusion constant [m^2]/[s]
raLTR = sqrt(2*raDiffusionD*raTauTR) # Mean free path [m]
raBTR = rhbar/(2*reConst*raLTR**2) # Magnetic transport field [T]
#############################
aResults = zeros((nTauPhi,nTauSO,nHoleDensities,2*nBValues)) # An empty array to contain the results of our calculation
for [iTauPhi, iTauSO] in iaTauPhiTauSOToCompute:
    rTauPhi = raTauPhiRange[iTauPhi] # tauPhi [s]
    rTauSO = raTauSORange[iTauSO] # tauSO [s]
    time0 = time()
    if verbosity:
            print('Rank '+str(rank)+': Evaluating iTauPhi,iTauSO = '+str(iTauPhi)+','+str(iTauSO))
    for iDensityIndex in iaHoleDensities:
        rTauTR = raTauTR[iDensityIndex]
        rBTR = raBTR[iDensityIndex]
        rTauTauPhi = rTauTR/rTauPhi # tauTR/tauPhi [unitless]
        rOmegaTau = rTauTR/rTauSO # tauTR/tauSO = Omega * tauTR [unitless]
        myraB, myraSig = evalModelPrediction(raB,rBTR,rTauTauPhi,rOmegaTau,nNCutoff,nQuadPoints,verbose=verbosity)
        aResults[iTauPhi,iTauSO,iDensityIndex,:] += myraSig
    time1 = time()
    if verbosity:
        print('Rank '+str(rank)+': Elapsed time: '+str(time1-time0)+' s.')
    fHandle = open(fOutputPrefix+str(rank)+fOutputSuffix,'wb')
    pickle.dump(aResults,fHandle) # Save the partial results to file
    fHandle.close()
