// Numerical evaluation of conductance shift for weak localization, following
// Eqns. (35) and (37) of [Glazov and Golub, Semiconductors 40, 1209 (2006)]
// Note: Assumes contribution from the cubic Dresselhaus spin-orbit coupling only.

// Jonathan E. Moussa
// May 2, 2018
// Sandia National Laboratories

// To compile, one may run something along the lines of:
// gcc -o conductanceShift conductanceShift.c -llapack -lblas

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>

// maximum number of terms to sum in the sigma formulas
#define NMAX 10000

// function declarations to silence compiler warnings
void dgesv_(int*,int*,double*,int*,int*,double*,int*,int*);
void dgemm_(char*,char*,int*,int*,int*,double*,double*,int*,double*,int*,double*,double*,int*);

// integration quadrature
void quadrature(double xmin, double xmax, int npt, double complex q, double* pt, double* wt, double complex* integrand)
{
  int i;
  double dx = (xmax - xmin)/(double)(npt-1);
  for(i=0 ; i<npt ; i++)
  {
    double x = xmin + dx*(double)i;

    if(x > 0.0)
    {
      double exp_x = exp(-x);
      pt[i] = x + log(1.0 + exp_x);
      wt[i] = dx/(1.0 + exp_x);
    }
    else
    {
      double exp_x = exp(x);
      pt[i] = log(exp_x + 1.0);
      wt[i] = dx*exp_x/(exp_x + 1.0);
    }

    integrand[i] = wt[i]*cexp(-q*pt[i]);
    wt[i] *= 2.0*pt[i];
    pt[i] *= pt[i];
  }
}

// fast, but possibly unstable version of the integrals
void fast_integral(int n, int m, double complex q, double complex* ans, double xmin, double xmax, int npt)
{
  int i,j;
  for(i=0 ; i<n ; i++) { ans[i] = 0.0; }

  // allocate & construct the quadrature
  double *pt = (double*)malloc(sizeof(double)*npt);
  double *wt = (double*)malloc(sizeof(double)*npt);
  double complex *integrand = (double complex*)malloc(sizeof(double complex)*npt);  
  quadrature(xmin,xmax,npt,q,pt,wt,integrand);

  // allocate & initialize the Laguerre functions
  double *L_old = (double*)malloc(sizeof(double)*npt);
  double *L = (double*)malloc(sizeof(double)*npt);
  double *L_new = (double*)malloc(sizeof(double)*npt);
  double L_wt = 1.0;
  for(i=1 ; i<=m ; i++) { L_wt /= sqrt((double)i); }
  for(i=0 ; i<npt ; i++) { L_old[i] = L_wt*exp(-pt[i]*0.5)*pow(pt[i],m*0.5); }
  L_wt = 1.0/sqrt(m+1.0);
  for(i=0 ; i<npt ; i++) { L[i] = L_wt*L_old[i]*(1.0+m-pt[i]); }

  // accumulate the first 2 integrals
  for(i=0 ; i<npt ; i++) { ans[0] += integrand[i]*L_old[i]; }
  if(n > 0) { for(i=0 ; i<npt ; i++) { ans[1] += integrand[i]*L[i]; } }

  // 3-term recurrence to construct the Laguerre functions
  for(i=1 ; i<n ; i++)
  {
    double wt1 = sqrt((i+1.0)/(i+m+1.0))/(i+1.0);
    double wt2 = wt1*sqrt((double)i/(double)(i+m));
    for(j=0 ; j<npt ; j++) { L_new[j] = (2*i+1+m-pt[j])*wt1*L[j] - (i+m)*wt2*L_old[j]; }
    double *L_temp = L_old;
    L_old = L;
    L = L_new;
    L_new = L_temp;

    for(j=0 ; j<npt ; j++) { ans[i+1] += integrand[j]*L[j]; }
  }

  // quadrature debugging information
//   FILE *quad_file = fopen("quad.out","w");
//   for(i=0 ; i<npt ; i++)
//   {
//     fprintf(quad_file,"%d %e %e %e %e\n",i,pt[i],L[i],creal(integrand[i]),cimag(integrand[i]));
//   }
//   fclose(quad_file);

  // free the quadrature & workspace
  free(pt); free(wt); free(integrand); free(L_old); free(L); free(L_new);
}

// slow, but stable version of the integrals
// NOTE: this is effectively deprecated because the 3-term recurrence appears to be stable w/o reorthogonalization
//       also, we are tailoring the integration quadrature for the primary integral rather than the inner product
void stable_integral(int n, int m, double complex q, double complex* ans, double xmin, double xmax, int npt)
{
  int i,j,k;
  for(i=0 ; i<n ; i++) { ans[i] = 0.0; }

  // allocate & construct the quadrature
  double *pt = (double*)malloc(sizeof(double)*npt);
  double *wt = (double*)malloc(sizeof(double)*npt);
  double complex *integrand = (double complex*)malloc(sizeof(double complex)*npt);  
  quadrature(xmin,xmax,npt,q,pt,wt,integrand);

  // allocate & initialize the Laguerre functions
  double **L = (double**)malloc(sizeof(double*)*(n+1));
  for(i=0 ; i<=n ; i++) { L[i] = (double*)malloc(sizeof(double)*npt); }

  double L_wt = 1.0;
  for(i=1 ; i<=m ; i++) { L_wt /= sqrt((double)i); }
  for(i=0 ; i<npt ; i++) { L[0][i] = L_wt*exp(-pt[i]*0.5)*pow(pt[i],m*0.5); }
  L_wt = 1.0/sqrt(m+1.0);
  for(i=0 ; i<npt ; i++) { L[1][i] = L_wt*L[0][i]*(1.0+m-pt[i]); }

  // 3-term recurrence to construct the Laguerre functions
  for(i=1 ; i<n ; i++)
  {
    double wt1 = sqrt((i+1.0)/(i+m+1.0))/(i+1.0);
    double wt2 = wt1*sqrt((double)i/(double)(i+m));
    for(j=0 ; j<npt ; j++) { L[i+1][j] = (2*i+1+m-pt[j])*wt1*L[i][j] - (i+m)*wt2*L[i-1][j]; }

    // explicit Gram-Schmidt orthogonalization
    for(j=0 ; j<=i ; j++)
    {
      double inner = 0.0;
      for(k=0 ; k<npt ; k++) { inner += wt[k]*L[j][k]*L[i+1][k]; }
      for(k=0 ; k<npt ; k++) { L[i+1][k] -= inner*L[j][k]; }
    }

    // explicit renormalization
    double norm = 0.0;
    for(j=0 ; j<npt ; j++) { norm += wt[j]*L[i+1][j]*L[i+1][j]; }
    norm = 1.0/sqrt(norm);
    for(j=0 ; j<npt ; j++) { L[i+1][j] *= norm; }
  }

  // accumulate the integrals at the end
  for(i=0 ; i<=n ; i++)
  { for(j=0 ; j<npt ; j++) { ans[i] += integrand[j]*L[i][j]; } }

  // free the quadrature & workspace
  for(i=0 ; i<=n ; i++) { free(L[i]); }
  free(pt); free(wt); free(integrand); free(L);
}

// common weight factor that appears in asymptotic matrix elements
double asym_wt(int n, int m)
{
  int i;
  double wt = 1.0;
  for(i=0 ; i<m ; i++) { wt *= (double)n/(double)(n+i+1); }
  return sqrt(wt);
}

// Fourier integral of Bessel function that appears in the asymptotic formulas
double complex asymptotic(int m, double a, double b)
{
  double complex c = csqrt(1.0 - cpow(a/(I + b),2));
  return (cpow(a/(1.0-I*b),m)/(1.0-I*b))/(c * cpow(1.0 + c,m));
}

struct data
{
  int n;
  double l_ratio;
  double l_ratio2;
  double omega_tau;
  double complex *integral0;
  double complex **integral;
  double complex **integral2;
};

double P(int n, struct data d)
{
  if(n<0) return 0.0;
  if(n>d.n) return creal(asymptotic(0,2.0*sqrt(n)/(d.l_ratio*d.l_ratio2),0.0))/d.l_ratio2;
  return d.l_ratio*creal(d.integral0[n]);
}
double Q(int n, struct data d)
{
  if(n<0) return 0.0;
  if(n>d.n) return asym_wt(n,1)*creal(asymptotic(1,2.0*sqrt(n)/d.l_ratio,0.0));
  return d.l_ratio*creal(d.integral[1][n]);
}
double R(int n, int m, struct data d)
{
  if(n<0) return 0.0;
  if(n>d.n) return (asym_wt(n,m)/sqrt(2.0))*cimag(asymptotic(m,2.0*sqrt(n)/d.l_ratio,2.0*d.omega_tau));
  return d.l_ratio*cimag(d.integral2[m][n])/sqrt(2.0);
}
double S(int n, int m, struct data d)
{
  if(n<0) return 0.0;
  if(n>d.n) return 0.5*asym_wt(n,m)*creal(asymptotic(m,2.0*sqrt(n)/d.l_ratio,0.0)
                                        - asymptotic(m,2.0*sqrt(n)/d.l_ratio,2.0*d.omega_tau));
  return d.l_ratio*creal(d.integral[m][n] - d.integral2[m][n])*0.5;
}
////
double PAsymptotic(int n, struct data d)
{
  if(n<0) return 0.0;
  return creal(asymptotic(0,2.0*sqrt(n)/(d.l_ratio*d.l_ratio2),0.0))/d.l_ratio2;
}
double QAsymptotic(int n, struct data d)
{
  if(n<0) return 0.0;
  return asym_wt(n,1)*creal(asymptotic(1,2.0*sqrt(n)/d.l_ratio,0.0));
}
double RAsymptotic(int n, int m, struct data d)
{
  if(n<0) return 0.0;
  return (asym_wt(n,m)/sqrt(2.0))*cimag(asymptotic(m,2.0*sqrt(n)/d.l_ratio,2.0*d.omega_tau));
}
double SAsymptotic(int n, int m, struct data d)
{
  if(n<0) return 0.0;
  return 0.5*asym_wt(n,m)*creal(asymptotic(m,2.0*sqrt(n)/d.l_ratio,0.0)
                                        - asymptotic(m,2.0*sqrt(n)/d.l_ratio,2.0*d.omega_tau));
}
////

// inverting I - mat
void inv_mat(double mat[9], double inv[9])
{
  int i, ipiv[3], info;
  double copy[9];
  for(i=0 ; i<9 ; i++) { copy[i] = -mat[i]; inv[i] = 0.0; }
  inv[0] = inv[4] = inv[8] = 1.0;
  copy[0] += 1.0; copy[4] += 1.0; copy[8] += 1.0;

  int n = 3;
  dgesv_(&n,&n,copy,&n,ipiv,inv,&n,&info);
}

double trace4(double mat1[9], double mat2[9], double mat3[9], double mat4[9])
{
  int i,j,k;
  char trans = 'N';
  int n = 3;
  double mat5[9], mat6[9], alpha = 1.0, beta = 0.0;

  for(i=0 ; i<9 ; i++) { mat5[i] = mat6[i] = 0.0; }
  dgemm_(&trans,&trans,&n,&n,&n,&alpha,mat1,&n,mat2,&n,&beta,mat5,&n);
  dgemm_(&trans,&trans,&n,&n,&n,&alpha,mat3,&n,mat4,&n,&beta,mat6,&n);

  double ans = 0.0;
  for(i=0 ; i<3 ; i++)
  for(j=0 ; j<3 ; j++)
  { ans += mat5[i+j*3]*mat6[j+i*3]; }

  return ans;
}

int main(int argc, char **argv)
{
  if(argc < 6) { printf("USAGE: <executable> <l_B/l> <l/{l~}> <Omega*tau> <max Laguerre order> <# of quadrature points>\n"); exit(1); }
  double l_ratio, l_ratio2, omega_tau, xmin, xmax;
  int n, npt;
  sscanf(argv[1],"%lf",&l_ratio);
  sscanf(argv[2],"%lf",&l_ratio2);
  sscanf(argv[3],"%lf",&omega_tau);
  sscanf(argv[4],"%d",&n);
  sscanf(argv[5],"%d",&npt);

  // rough heuristic values of xmin & xmax
  double scale = 1.0;
  if(1.0/(double)n < scale) { scale = 1.0/(double)n; }
  if(1.0/l_ratio < scale) { scale = 1.0/l_ratio; }
  double log_eps = npt * sqrt(scale);
  xmin = -log_eps;
  xmax = sqrt(2.0*log_eps + pow(l_ratio,2)) - l_ratio;

  // generate all the raw integrals using the fast algorithm
  int i,j,k;
  double complex *integral0 = (double complex*)malloc(sizeof(double complex)*(n+1));
  double complex **integral = (double complex**)malloc(sizeof(double complex*)*8); // w/o oscillations
  double complex **integral2 = (double complex**)malloc(sizeof(double complex*)*8); // w/ oscillations
  for(i=0 ; i<8 ; i++)
  {
    integral[i] = (double complex*)malloc(sizeof(double complex)*(n+1));
    integral2[i] = (double complex*)malloc(sizeof(double complex)*(n+1));
  }

  fast_integral(n,0,l_ratio*l_ratio2,integral0,xmin,xmax,npt);
  for(i=0 ; i<8 ; i++)
  {
   fast_integral(n,i,l_ratio,integral[i],xmin,xmax,npt);
   fast_integral(n,i,l_ratio - 2.0*l_ratio*omega_tau*I,integral2[i],xmin,xmax,npt);
  }

//   FILE *conv_file = fopen("conv.out","w");

  // calculate & print term-by-term contributions to sigma_a & sigma_b
  struct data d; d.n = n; d.l_ratio = l_ratio; d.l_ratio2 = l_ratio2; d.omega_tau = omega_tau;
  d.integral0 = integral0; d.integral = integral; d.integral2 = integral2;
  double A[9], K[9], ImAinv[9], M1[9], M2[9];
  double a = 0.0, b = 0.0;
  double awt = -1.0/(2.0*M_PI*M_PI*l_ratio*l_ratio);
  double bwt = 1.0/(4.0*M_PI*M_PI*l_ratio*l_ratio);
  for(i=0 ; i<NMAX ; i++)
  {
    // accumulate all non-matrix contributions to sigma_a & sigma_b
    double da = -pow(P(i,d),3)/(1.0 - P(i,d));
    double db = -pow(Q(i,d),2)*( P(i,d)/(1.0 - P(i,d)) + P(i+1,d)/(1.0 - P(i+1,d)) );

    // construct A_i, A_{i+1}, & K matrices
    double A[9], A2[9], K[9], KT[9];
    A[0 + 0*3] = P(i-6,d) - S(i-6,0,d);
    A[0 + 1*3] = R(i-6,3,d);
    A[0 + 2*3] = S(i-6,6,d);
    A[1 + 0*3] = R(i-3,3,d);
    A[1 + 1*3] = P(i-3,d) - 2.0*S(i-3,0,d);
    A[1 + 2*3] = R(i-3,3,d);
    A[2 + 0*3] = S(i,6,d);
    A[2 + 1*3] = R(i,3,d);
    A[2 + 2*3] = P(i,d) - S(i,0,d);
    i++;
    A2[0 + 0*3] = P(i-6,d) - S(i-6,0,d);
    A2[0 + 1*3] = R(i-6,3,d);
    A2[0 + 2*3] = S(i-6,6,d);
    A2[1 + 0*3] = R(i-3,3,d);
    A2[1 + 1*3] = P(i-3,d) - 2.0*S(i-3,0,d);
    A2[1 + 2*3] = R(i-3,3,d);
    A2[2 + 0*3] = S(i,6,d);
    A2[2 + 1*3] = R(i,3,d);
    A2[2 + 2*3] = P(i,d) - S(i,0,d);
    i--;
    K[0 + 0*3] = Q(i-6,d) - S(i-6,1,d);
    K[0 + 1*3] = -R(i-5,2,d);
    K[0 + 2*3] = -S(i-5,5,d);
    K[1 + 0*3] = R(i-6,4,d);
    K[1 + 1*3] = Q(i-3,d) - 2.0*S(i-3,1,d);
    K[1 + 2*3] = -R(i-2,2,d);
    K[2 + 0*3] = S(i-6,7,d);
    K[2 + 1*3] = R(i-3,4,d);
    K[2 + 2*3] = Q(i,d) - S(i,1,d);
    for(j=0 ; j<3 ; j++) for(k=0 ; k<3 ; k++) { KT[j + k*3] = K[k + j*3]; }

    // invert I-A_i & I-A_{i+1}
    double ImAinv[9], ImA2inv[9];
    inv_mat(A,ImAinv);
    inv_mat(A2,ImA2inv);

    // accumulate matrix contributions to sigma_a & sigma_b
    da += trace4(A,A,A,ImAinv);
    db += trace4(K,KT,A,ImAinv) + trace4(KT,K,A2,ImA2inv);

    a += da; b += db;
//     fprintf(conv_file,"%d %16.16e %16.16e\n",i,awt*a,bwt*b);
    // Term-by-term checking of the asymptotic expansion against the numerical integral:
    //     double PA, QA, SA, RA, PN, QN, SN, RN;
    //     PN = P(i,d);
    //     QN = Q(i,d);
    //     SN = S(i,3,d);
    //     RN = R(i,3,d);
    //     PA = PAsymptotic(i,d);
    //     QA = QAsymptotic(i,d);
    //     SA = SAsymptotic(i,3,d);
    //     RA = RAsymptotic(i,3,d);
    //     fprintf(conv_file,"%d %16.16e %16.16e %16.16e %16.16e %16.16e %16.16e %16.16e %16.16e %16.16e %16.16e\n",i,awt*a,bwt*b,PA,QA,SA,RA,PN,QN,SN,RN);
  }
//   fclose(conv_file);

//  printf("sigma_a = %16.16e\n",awt*a);
//  printf("sigma_b = %16.16e\n",bwt*b);

  // printout to reproduce Fig. 4
  printf("%e %e %e\n",2.0/pow(l_ratio,2),awt*a+bwt*b,omega_tau);

  for(i=0 ; i<8 ; i++) { free(integral[i]); free(integral2[i]); }
  free(integral2); free(integral); free(integral0);
  return 1;
}
